#include "book.h"
#include <vector>
#include <cstdlib>
#include <time.h>
#include <string>
#include <algorithm>
#include <iostream>
#ifdef _MSC_VER // Windows
#include <process.h>
#else
#include <sys/types.h>
#include <unistd.h>
#endif
#include "utils.h"
#include "pgn.h"


namespace polyglot {

	book::book() {
		this->open(fileName, std::ifstream::in | std::ifstream::binary);
		this->seekg(0, std::ios::end);
		count = size_t(this->tellg() / sizeof(Entry));
#ifdef _MSC_VER // Windows
		srand(uint32_t(time(NULL)*_getpid()));
#else
        srand(uint32_t(time(NULL)*getpid()));
#endif
	}

	book::book(const std::string& filename)
	{
		fileName = filename;
		this->open(fileName, std::ifstream::in | std::ifstream::binary);
		this->seekg(0, std::ios::end);
		count = size_t(this->tellg() / sizeof(Entry));
#ifdef _MSC_VER // Windows
		srand(uint32_t(time(NULL)*_getpid()));
#else
        srand(uint32_t(time(NULL)*getpid()));
#endif
	}


	book::~book()
	{
		if (is_open()) close();
	}

	Entry book::read() {
		Entry result;
		result.key = 0;
		result.move = 0;
		result.weight = 0;
		for (int i = 0; i < 8; ++i) result.key = (result.key << 8) + std::ifstream::get();
		for (int i = 0; i < 2; ++i) result.move = (result.move << 8) + std::ifstream::get();
		for (int i = 0; i < 2; ++i) result.weight = (result.weight << 8) + std::ifstream::get();
		return result;
	}

	Move book::probe(position& pos, bool pickBest, ValuatedMove * moves, int moveCount) {
		if (!is_open()) return MOVE_NONE;
		//Make a binary search to find the right Entry
		size_t low = 0;
		size_t high = count - 1;
		Entry entry;
		size_t searchPoint = 0;
		while (low < high && good()) {
			searchPoint = low + (high - low) / 2;
			seekg(searchPoint * sizeof(Entry), ios_base::beg);
			entry = read();
			if (pos.GetHash() == entry.key) break;
			else if (pos.GetHash() < entry.key) high = searchPoint; else low = searchPoint + 1;
		}
#pragma warning(suppress: 6001)
		if (entry.key != pos.GetHash()) return MOVE_NONE;
		std::vector<Entry> entries;
		entries.push_back(entry);
		size_t searchPointStore = searchPoint;
		while (searchPoint > 0 && good()) {
			searchPoint--;
			seekg(searchPoint * sizeof(Entry), ios_base::beg);
			entry = read();
			if (pos.GetHash() != entry.key) break; else entries.push_back(entry);
		}
		searchPoint = searchPointStore;
		while (searchPoint < (count - 1) && good()) {
			searchPoint++;
			seekg(searchPoint * sizeof(Entry), ios_base::beg);
			entry = read();
			if (pos.GetHash() != entry.key) break; else entries.push_back(entry);
		}
		Entry best = entries[0];
		uint32_t sum = best.weight;
		for (size_t i = 1; i < entries.size(); ++i) {
			Entry e = entries[i];
			if (e.weight > best.weight) best = e;
			sum += e.weight;
		}
		Move move = best.move;
		if (!pickBest && entries.size() > 0) {
			uint32_t indx = rand() % sum;
			sum = 0;
			for (size_t i = 0; i < entries.size(); ++i) {
				Entry e = entries[i];
				sum += e.weight;
				if (sum > indx) {
					move = e.move;
					break;
				}
			}
		}
		//Now move contains move in polyglot representation
		int pt = (move >> 12) & 7;
		if (pt)
			move = createMove<PROMOTION>(from(move), to(move), PieceType(4 - pt));
		else {
			//Castling is stored as king captures rook
			if (pos.GetSideToMove() == WHITE && from(move) == InitialKingSquare[WHITE]) {
				if (to(move) == InitialRookSquare[0] && pos.CastlingAllowed(W0_0)) {
					move = createMove<CASTLING>(InitialKingSquare[WHITE], G1);
				}
				else if (to(move) == InitialRookSquare[1] && pos.CastlingAllowed(W0_0_0)) {
					move = createMove<CASTLING>(InitialKingSquare[WHITE], C1);
				}
			}
			else if (pos.GetSideToMove() == BLACK && from(move) == InitialKingSquare[BLACK]) {
				if (to(move) == InitialRookSquare[2] && pos.CastlingAllowed(B0_0)) {
					move = createMove<CASTLING>(InitialKingSquare[BLACK], G8);
				}
				else if (to(move) == InitialRookSquare[3] && pos.CastlingAllowed(B0_0_0)) {
					move = createMove<CASTLING>(InitialKingSquare[BLACK], C8);
				}
			}
		}
		//Now check if polyglot move is part of generated move list
		move &= 0x3FFF;
		for (int i = 0; i < moveCount; i++) {
			if ((moves[i].move & 0x3FFF) == move) return moves[i].move;
		}
		return MOVE_NONE;
	}

}
